﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.Serialization;
using System.Text.RegularExpressions;
#if SPYRO_LINQ2
using Spyro.Linq2;
#else
using System.Linq;
#endif

namespace $rootnamespace$.LineaComandos
{
    internal partial class Argumentos
    {
        public static Argumentos Tratar<T>(string[] p_args, T v_definiciones, Action<ErrorArgumentosException> p_mostrarErrorArgumentos)
        {
            Argumentos v_argumentos;
            try
            {
                var v_descripcion = Utilidades.ObtenerValorAtributo<AssemblyDescriptionAttribute>(p_atributo => p_atributo.Description);
                var v_nombrePrograma = Utilidades.ObtenerValorAtributo<AssemblyProductAttribute>(p_atributo => p_atributo.Product);

                v_argumentos = Argumentos.Crear(v_descripcion, v_nombrePrograma, p_args, v_definiciones);
            }
            catch (ErrorArgumentosException e)
            {
                Trace.WriteLine(e.Message);
                p_mostrarErrorArgumentos(e);
                Environment.Exit(1);
                return null;
            }
            catch (Exception e)
            {
                Trace.WriteLine(e.Message);
                Console.WriteLine(e.Message);
                Environment.Exit(1); // Parámetros incorrectos.
                return null;
            }
            return v_argumentos;
        }
        
        public static Argumentos Tratar<T>(string[] p_args, T v_definiciones) 
        {
            return Tratar<T>(p_args, v_definiciones, p_excepcion => 
            {
                Console.WriteLine(p_excepcion.Message);
                Console.WriteLine();
                Utilidades.EscribeUso(p_excepcion);
            });
        }
        
        private Dictionary<string, object> valores;
        public string Uso { get; private set; }

        internal Argumentos(Dictionary<string, object> p_valores, string p_uso)
        {
            this.valores = p_valores;
            this.Uso = p_uso;
        }

        public object this[string p_parametro]
        {
            get
            {
                object v_valor;
                if (this.valores.TryGetValue(p_parametro, out v_valor))
                    return v_valor;
                else
                    return null;
            }
        }

        public static Argumentos Crear<T>(string[] p_args, T p_definiciones)
        {
            return Crear(null, null, p_args, p_definiciones);
        }

        public static Argumentos Crear<T>(string p_descripcion, string p_programa, string[] p_args, T p_definiciones)
        {
            var v_definiciones = p_definiciones.GetType().GetProperties().
                Select(p => new { Parametro = p.Name, DescriptorDefinicion = (IDescriptorArgumento)p.GetValue(p_definiciones, null) }).
                ToDictionary(p => p.Parametro, p => ObtenerDefinicion(p.Parametro, p.DescriptorDefinicion), StringComparer.InvariantCultureIgnoreCase);

            var v_valores = p_args
                .SelectMany(p_arg =>
                {
                    string v_parametro, v_valor;

                    if (p_arg.StartsWith("-", StringComparison.Ordinal))
                    {
                        int i = p_arg.IndexOf('=');
                        if (i > 0)
                        {
                            v_parametro = p_arg.Substring(1, i - 1);
                            v_valor = p_arg.Substring(i + 1);
                        }
                        else {
                            v_parametro = p_arg.Substring(1);
                            v_valor = "";
                        }
                        return new[] { new { Parametro = v_parametro, Valor = v_valor } };
                    }
                    else {
                        if (v_definiciones.ContainsKey("usuario") &&
                            v_definiciones.ContainsKey("password") &&
                            v_definiciones.ContainsKey("bd"))
                        {
                            string v_arg = p_arg;
                            string v_usuario, v_password = null, v_bd = null;
                            int i = v_arg.IndexOf('/');
                            if (i > 0)
                            {
                                v_usuario = v_arg.Substring(0, i);
                                v_arg = v_arg.Substring(i + 1);
                                i = v_arg.IndexOf('@');
                                if (i > 0)
                                {
                                    v_password = v_arg.Substring(0, i);
                                    v_bd = v_arg.Substring(i + 1);
                                }
                                else // no hay bd (@)
                                    v_password = v_arg;
                            }
                            else { // no hay password (/)
                                i = v_arg.IndexOf('@');
                                if (i > 0)
                                {
                                    v_usuario = v_arg.Substring(0, i);
                                    v_bd = v_arg.Substring(i + 1);
                                }
                                else // no hay bd (@)
                                    v_usuario = v_arg;
                            }

                            return new[]
                                {
                                    new { Parametro = "usuario", Valor = v_usuario } ,
                                    new { Parametro = "password", Valor = v_password } ,
                                    new { Parametro = "bd", Valor = v_bd } ,
                                };
                        }
                        else
                            return new[] { new { Parametro = p_arg, Valor = default(string) } };
                    }
                })
                .Where(p => !string.IsNullOrEmpty(p.Parametro))
                .Select(p =>
                {
                    Definicion v_definicion;
                    var v_valido = false;
                    object v_valor = null;
                    string v_error = null;
                    if (v_definiciones.TryGetValue(p.Parametro, out v_definicion))
                    {
                        v_valido = v_definicion.Validar(p.Valor, out v_error);
                        if (!string.IsNullOrEmpty(v_error) && !v_error.EndsWith("."))
                        {
                            if (v_error.Length > 1)
                                v_error = v_error.Substring(0, 1).ToLower() + v_error.Substring(1);
                            v_error += ".";
                        }
                        if (v_valido)
                            v_valor = v_definicion.Convertir(p.Valor);
                    }
                    return new { Parametro = p.Parametro, Definicion = v_definicion, Valido = v_valido, Valor = v_valor, Error = v_error };
                })
                .ToArray();

            var v_diccionarioValores = v_valores
                .Where(p => p.Valor != null)
                .ToDictionary(s => s.Parametro, s => s.Valor, StringComparer.InvariantCultureIgnoreCase);

            var v_mensaje = "";

            var v_sinDefinicion = v_valores
                .Where(p => p.Definicion == null)
                .ToArray();

            if (v_sinDefinicion.Length > 1)
                v_mensaje = string.Format("{0} no son parámetros válidos.", v_sinDefinicion.Select(p => string.Format("'{0}'", p.Parametro)).Aggregate((a, b) => a + ", " + b));
            else if (v_sinDefinicion.Length == 1)
                v_mensaje = string.Format("'{0}' no es un parámetro válido.", v_sinDefinicion.First().Parametro);

            var v_noValidos = v_valores
               .Where(p => p.Definicion != null && !p.Valido)
               .ToArray();

            if (v_noValidos.Length > 0)
            {
                if (!string.IsNullOrEmpty(v_mensaje))
                    v_mensaje += "\n";
                v_mensaje = string.Format("{0} ", v_noValidos.Select(p_noValido => string.Format("'{0}' {1}", p_noValido.Definicion.Parametro, p_noValido.Error)).Aggregate((a, b) => a + "\n" + b));
            }

            var v_faltan = v_definiciones
                .Where(d => !d.Value.Opcional)
                .Where(d => !v_diccionarioValores.ContainsKey(d.Key))
                .ToArray();

            if (v_faltan.Length > 1)
            {
                if (!string.IsNullOrEmpty(v_mensaje))
                    v_mensaje += "\n";
                v_mensaje += string.Format("{0} son obligatorios.", v_faltan.Select(p => string.Format("'{0}'", p.Key)).Aggregate((a, b) => a + ", " + b));
            }
            else if (v_faltan.Length == 1)
            {
                if (!string.IsNullOrEmpty(v_mensaje))
                    v_mensaje += "\n";
                v_mensaje += string.Format("'{0}' es obligatorio.", v_faltan.First().Key);
            }

            var v_uso = "";
            if (!string.IsNullOrEmpty(p_descripcion))
                v_uso = p_descripcion;

            if (!string.IsNullOrEmpty(p_programa))
            {
                if (!string.IsNullOrEmpty(v_uso))
                    v_uso += "\n\n";
                v_uso += string.Format("{0} ", p_programa);
            }

            v_uso += v_definiciones.Select(d => string.Format(d.Value.Opcional ? "[-{0}{1}]" : "-{0}{1}", d.Key, (!string.IsNullOrEmpty(d.Value.Valor) ? "=" : "") + d.Value.Valor)).Aggregate((a, b) => a + " " + b);
            var v_espaciosTabulador = v_definiciones.Max(d => d.Key.Length) + 1;
            v_uso += string.Format("\n\n{0}", v_definiciones.Select(d => d.Value.ObtenerDescripcionUso(v_espaciosTabulador - d.Key.Length)).Aggregate((a, b) => a + "\n" + b));

            if (!string.IsNullOrEmpty(v_mensaje))
                throw new ErrorArgumentosException(v_mensaje, v_uso);

            return new Argumentos(v_diccionarioValores, v_uso);
        }

        private static Definicion ObtenerDefinicion(string p_parametro, IDescriptorArgumento p_descriptor)
        {
            var v_opcional = false;
            var v_descripcion = "";
            var v_descripcionValor = "";
            var v_validadores = new List<Validador>();
            DelegadoConversor v_conversor = null;

            var v_descriptor = p_descriptor;
            while (v_descriptor != null)
            {
                if (v_descriptor is DescriptorInt)
                {
                    v_validadores.Add(new Validador(p_valor => { int i; return int.TryParse((string)p_valor, out i); }, "No es un valor numérico válido"));
                    v_conversor = p_valor => int.Parse(p_valor);
                }
                else if (v_descriptor is DescriptorLong)
                {
                    v_validadores.Add(new Validador(p_valor => { long l; return long.TryParse((string)p_valor, out l); }, "No es un valor numérico válido"));
                    v_conversor = p_valor => long.Parse(p_valor);
                }
                else if (v_descriptor is DescriptorDecimal)
                {
                    v_validadores.Add(new Validador(p_valor => { decimal d; return decimal.TryParse((string)p_valor, out d); }, "No es un valor numérico válido"));
                    v_conversor = p_valor => decimal.Parse(p_valor);
                }
                else if (v_descriptor is DescriptorDouble)
                {
                    v_validadores.Add(new Validador(p_valor => { double d; return double.TryParse((string)p_valor, out d); }, "No es un valor numérico válido"));
                    v_conversor = p_valor => double.Parse(p_valor);
                }
                else if (v_descriptor is DescriptorFormatoString)
                {
                    v_conversor = p_valor => p_valor;
                }
                else if (v_descriptor is DescriptorDate)
                {
                    v_validadores.Add(new Validador(p_valor => { DateTime d; return DateTime.TryParse((string)p_valor, out d); }, "No es una fecha válida"));
                    v_conversor = p_valor => DateTime.Parse(p_valor);
                }
                else if (v_descriptor is IDescriptorOpcional)
                {
                    v_opcional = true;
                }
                else if (v_descriptor is DescriptorDescripcion)
                {
                    v_descripcion = ((DescriptorDescripcion)v_descriptor).Descripcion;
                }
                else if (v_descriptor is DescriptorDescripcionValor)
                {
                    v_descripcionValor = ((DescriptorDescripcionValor)v_descriptor).DescripcionValor;
                }
                else if (v_descriptor is IDescriptorValidador)
                {
                    v_validadores.Add(new Validador(((IDescriptorValidador)v_descriptor).Validar, ((IDescriptorValidador)v_descriptor).MensajeError));
                }
                else
                    throw new NotSupportedException();

                if (v_descriptor is IDescriptorTienePadre)
                    v_descriptor = ((IDescriptorTienePadre)v_descriptor).Padre;
                else
                    v_descriptor = null;
            }

            var v_definicion = new Definicion(p_parametro, p_descriptor, v_opcional, v_descripcion, v_descripcionValor, v_validadores.ToArray(), v_conversor);
            return v_definicion;
        }

        public static DescriptorFormatoString String()
        {
            return new DescriptorFormatoString();
        }

        public static DescriptorInt Int()
        {
            return new DescriptorInt();
        }

        public static DescriptorLong Long()
        {
            return new DescriptorLong();
        }

        public static DescriptorDecimal Decimal()
        {
            return new DescriptorDecimal();
        }

        public static DescriptorDouble Double()
        {
            return new DescriptorDouble();
        }

        public static DescriptorInt Int32()
        {
            return new DescriptorInt();
        }

        public static DescriptorInt Int64()
        {
            return new DescriptorInt();
        }

        public static DescriptorDate Date()
        {
            return new DescriptorDate();
        }

        private delegate object DelegadoConversor(string p_valor);

        private class Definicion
        {
            public string Parametro { get; private set; }
            public IDescriptorArgumento Descriptor { get; private set; }
            public bool Opcional { get; private set; }
            public string Descripcion { get; private set; }
            public string Valor { get; private set; }
            public Validador[] Validadores { get; private set; }
            public DelegadoConversor Conversor { get; private set; }

            public bool Validar(string p_valor, out string p_mensaje)
            {
                foreach (var v_validador in this.Validadores)
                {
                    if (!v_validador.Validar(p_valor))
                    {
                        p_mensaje = v_validador.MensajeError;
                        return false;
                    }
                }
                p_mensaje = null;
                return true;
            }

            public object Convertir(string p_valor) { return Conversor(p_valor); }

            public Definicion(string p_parametro, IDescriptorArgumento p_descriptor, bool p_opcional, string p_descripcion, string p_descripcionValor,
                Validador[] p_validadores, DelegadoConversor p_conversor)
            {
                this.Parametro = p_parametro;
                this.Descriptor = p_descriptor;
                this.Opcional = p_opcional;
                this.Descripcion = p_descripcion;
                this.Valor = p_descripcionValor;
                this.Validadores = p_validadores;
                this.Conversor = p_conversor;
            }

            public string ObtenerDescripcionUso(int p_espaciosTabulador)
            {
                var v_espacios = new string(' ', p_espaciosTabulador);
                return string.Format("  {0}{2}{1}", this.Parametro, this.Descripcion, v_espacios);
            }
        }
    }

    internal interface IDescriptorArgumento
    {
    }

    internal interface IDescriptorDefinicion<T> : IDescriptorArgumento
    {
    }

    internal class DescriptorFormatoString : IDescriptorDefinicion<string>
    {
        public DescriptorOpcional<string> Opcional()
        {
            return new DescriptorOpcional<string>(this);
        }
    }

    internal interface IDescriptorTienePadre
    {
        IDescriptorArgumento Padre { get; }
    }

    internal interface IDescriptorOpcional : IDescriptorTienePadre
    {
    }

    internal class DescriptorInt : IDescriptorDefinicion<int>
    {
        public DescriptorOpcionalNulable<int> Opcional()
        {
            return new DescriptorOpcionalNulable<int>(this);
        }
    }

    internal class DescriptorLong : IDescriptorDefinicion<long>
    {
        public DescriptorOpcionalNulable<long> Opcional()
        {
            return new DescriptorOpcionalNulable<long>(this);
        }
    }

    internal class DescriptorDecimal : IDescriptorDefinicion<decimal>
    {
        public DescriptorOpcionalNulable<decimal> Opcional()
        {
            return new DescriptorOpcionalNulable<decimal>(this);
        }
    }

    internal class DescriptorDouble : IDescriptorDefinicion<double>
    {
        public DescriptorOpcionalNulable<double> Opcional()
        {
            return new DescriptorOpcionalNulable<double>(this);
        }
    }

    internal class DescriptorDate : IDescriptorDefinicion<DateTime>
    {
        public DescriptorOpcionalNulable<DateTime> Opcional()
        {
            return new DescriptorOpcionalNulable<DateTime>(this);
        }
    }

    internal class DescriptorOpcional<T> : IDescriptorDefinicion<T>, IDescriptorOpcional, IDescriptorTienePadre where T : class
    {
        public IDescriptorArgumento Padre { get; private set; }

        public DescriptorOpcional(IDescriptorDefinicion<T> padre)
        {
            this.Padre = padre;
        }
    }

    internal class DescriptorOpcionalNulable<T> : IDescriptorDefinicion<T?>, IDescriptorOpcional, IDescriptorTienePadre where T : struct
    {
        public IDescriptorArgumento Padre { get; private set; }

        public DescriptorOpcionalNulable(IDescriptorDefinicion<T> padre)
        {
            this.Padre = padre;
        }
    }

    internal delegate bool DelegadoValidador(object p_valor);

    internal class Validador
    {
        public DelegadoValidador Validar { get; private set; }

        public string MensajeError { get; private set; }

        public Validador(DelegadoValidador p_validar, string p_mensajeError)
        {
            this.Validar = p_validar;
            this.MensajeError = p_mensajeError;
        }
    }

    internal interface IDescriptorValidador : IDescriptorTienePadre
    {
        bool Validar(object p_valor);

        string MensajeError { get; }
    }

    internal delegate bool DelegadoDescriptorValidacion<T>(T p_valor);

    internal class DescriptorValidacion<T> : IDescriptorDefinicion<T>, IDescriptorValidador, IDescriptorTienePadre
    {
        public IDescriptorArgumento Padre { get; private set; }

        private readonly DelegadoDescriptorValidacion<T> func;
        public string MensajeError { get; private set; }

        public DescriptorValidacion(IDescriptorArgumento p_padre, DelegadoDescriptorValidacion<T> p_func, string p_mensaje)
        {
            this.Padre = p_padre;
            this.func = p_func;
            this.MensajeError = p_mensaje;
        }

        public bool Validar(object p_valor)
        {
            T v_valor = (T)Convert.ChangeType(p_valor, typeof(T));
            return this.func(v_valor);
        }
    }

    internal class DescriptorDescripcion : IDescriptorArgumento, IDescriptorTienePadre
    {
        public IDescriptorArgumento Padre { get; private set; }

        public string Descripcion { get; private set; }

        public DescriptorDescripcion(IDescriptorArgumento padre, string descripcion)
        {
            this.Padre = padre;
            this.Descripcion = descripcion;
        }
    }

    internal class DescriptorDescripcionValor : IDescriptorArgumento, IDescriptorTienePadre
    {
        public IDescriptorArgumento Padre { get; private set; }

        public string DescripcionValor { get; private set; }

        public DescriptorDescripcionValor(IDescriptorArgumento padre, string descripcionValor)
        {
            this.Padre = padre;
            this.DescripcionValor = descripcionValor;
        }
    }

    internal static class UtilidadesArgumentos
    {
        public static IDescriptorArgumento Descripcion(this IDescriptorArgumento p_descriptor, string p_descripcion)
        {
            return new DescriptorDescripcion(p_descriptor, p_descripcion);
        }

        public static IDescriptorArgumento Valor(this IDescriptorArgumento p_descriptor, string p_valor)
        {
            return new DescriptorDescripcionValor(p_descriptor, p_valor);
        }

        public static IDescriptorDefinicion<T> Valida<T>(this IDescriptorDefinicion<T> p_descriptor, DelegadoDescriptorValidacion<T> p_validacion, string p_mensaje)
        {
            return new DescriptorValidacion<T>(p_descriptor, p_validacion, p_mensaje);
        }

        public static IDescriptorDefinicion<string> Valida(this IDescriptorDefinicion<string> p_descriptor, string p_patron, string p_mensaje)
        {
            return new DescriptorValidacion<string>(p_descriptor, p_valor => Regex.IsMatch(p_valor, p_patron), p_mensaje);
        }        
    }

    [Serializable]
    public class ErrorArgumentosException : Exception
    {
        public string Uso { get; private set; }

        public ErrorArgumentosException() { }
        public ErrorArgumentosException(string message) : base(message) { }
        public ErrorArgumentosException(string message, Exception inner) : base(message, inner) { }
        protected ErrorArgumentosException(SerializationInfo info, StreamingContext context) : base(info, context) { }

        public ErrorArgumentosException(string message, string uso)
            : base(message)
        {
            this.Uso = uso;
        }
    }
}